<?php
/*
 * Enrober v1.2.0 - PHP theming engine
 *
 * Copyright 2011  Peter Johnson
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @package enrober
 * @author Peter Johnson
 * @copyright Copyright (c) 2011, Peter Johnson
 * @version 1.2.0
 * @license http://www.apache.org/licenses/LICENSE-2.0 Apache License, Version 2.0
 * @link http://www.uselesscode.org/php/enrober/
 */

/**
 */

// Try to load customized config
if (file_exists(__DIR__ . '/cfg/config.php')) {
	require_once(__DIR__ . '/cfg/config.php');
} else {
	// if it fails fallback on default
 	require_once(__DIR__ . '/cfg/config.default.php');
}



/**
 * @package enrober
 */
/**
 * @ignore
 * Filter the body and template after they have been combined
 * @set add_filter()
 */
define('ENROBER_FILTER_ALL', 0);
/**
 * @ignore
 * Filter the header data after tags have been injected
 * @set add_filter()
 */
define('ENROBER_FILTER_HEAD', 1);
/**
 * @ignore
 * Filter just the body of the calling page
 * @set add_filter()
 */
define('ENROBER_FILTER_BODY', 2);
/**
 * @ignore
 * Filter the footer data after tags have been injected
 * @set add_filter()
 */
define('ENROBER_FILTER_FOOT', 3);
/**
 * @ignore
 * Filter the whole page after it has been combined
 * @set add_filter()
 */
define('ENROBER_FILTER_EVERYTHING ', 4);

/**
 * @ignore
 * Filters to be applied to XHR requests before they are returned.
 * @set add_filter()
 */
define('ENROBER_FILTER_XHR', 5);
/**
 * @ignore
 * Filters header, allways applied at request time, even on cached items.
 * @set add_filter()
 */
define('ENROBER_FILTER_ALWAYS_HEAD', 6);
/**
 * @ignore
 * Filters content, allways applied at request time, even on cached items.
 * @set add_filter()
 */
define('ENROBER_FILTER_ALWAYS_BODY', 7);
/**
 * @ignore
 * Filters footer, allways applied at request time, even on cached items.
 * @set add_filter()
 */
define('ENROBER_FILTER_ALWAYS_FOOT', 8);

/**
 * @package enrober
 * @subpackage classes
 */
class Enrober {
	/**
	 * Once an instance of the Enrober class is created using the singleton() method,
	 * it is stored in $instance.
	 * @private
	 */
	private static $instance;

	// Ini to null so default can be set based on config CONSTANTS
	// or overridden by methods or other logic in go()
	private $cache_pages = null; // should caching be used?
	private $cache_valid = null; // is the cache file for the current page still valid

	/**
	 * @var string The currently selected theme
	 */
	private $theme_name = null;

	/**
	 * @var string The title of the page
	 */
	private $page_title = null;

	/**
	 * Array of CSS files to link
	 */
	private $css_stack = array();

	/**
	 * Array of JavaScript files to link
	 */
	private $js_stack = array();

	/**
	 * Array of filters to apply to the body of the page.
	 */
	private $filter_body_stack = array();

	/**
	 * Array of filters to apply to the header.
	 */
	private $filter_head_stack = array();

	/**
	 * Array of filters to apply to the footer.
	 */
	private $filter_foot_stack = array();

	/**
	 * Array of filters to apply to the entire page after it's been buffered.
	 */
	private $filter_everything_stack = array();

	/**
	 * Array of filters to apply to XHR requests.
	 */
	private $filter_xhr_stack = array();

	/**
	 * Array of filters to apply just before the page is served, even on cached files.
	 */
	private $filter_always_head_stack = array();

	/**
	 * Array of filters to apply just before the page is served, even on cached files.
	 */
	private $filter_always_body_stack = array();

	/**
	 * Array of filters to apply just before the page is served, even on cached files.
	 */
	private $filter_always_foot_stack = array();

	/**
	 * Array of tags that will be processed in the template header.
	 */
	private $tag_stack_head = array();

	/**
	 * Array of tags that will be processed in the template footer.
	 */
	private $tag_stack_foot = array();

	/**
	 * Array of modifiers that will be used in generating the cache file filename.
	 */
	private $modifiers = array();

	/**
	 * String of keywords for a <meta> tag
	 */
	private $keywords = '';

	/**
	 * String of the description for a <meta> tag
	 */
	private $desc = '';

	 /**
	 * The canonical URL for this page
	 */
	private $url = null;

	/**
	 * Are we currently in maintenance mode?
	 */
	private $is_maintenance_mode = null;

	/**
	 * Is the current user the maintenance user?
	 */
	private $is_maintenance_user = null;

	/**
	 * Is this page allowed during maintenance mode?
	 */
	private $page_allowed_during_maintenance = false; // guilty until proven innocent

	/**
	 * Is the User Agent mobile, or are we forcing it to
	 * the mobile or non-mobile version.
	 *
	 * Null means go() will auto-detect, if true or false it was manually set via 
	 * force_mobile()
	 */
	private $mobile = null;

	/**
	 * @ignore
	 */
	private function __construct() {
		$this->is_maintenance_mode = (defined('ENROBER_MAINTENANCE_DONE_AT') && ENROBER_MAINTENANCE_DONE_AT !== null) ? true : false;
		$this->is_maintenance_user = (isset($_SERVER['REMOTE_ADDR']) && $_SERVER['REMOTE_ADDR'] === ENROBER_MAINTENANCE_IP) ? true : false;
		if ($this->is_maintenance_mode) {
			$this->page_allowed_during_maintenance = $this->allowed_during_maintenance($_SERVER['SCRIPT_NAME']);
		} else {
			$this->page_allowed_during_maintenance = true;
		}
	}

	/**
	 * Creates the $enrober singleton object.
	 * Automatically called when enrober.php is included.
	 * The created object is assigned to the $enrober variable.
	 * Because Enrober is a singleton, further calls to singleton()
	 * will result in a reference to the original object instead
	 * of a new instance.
	 * @returns object
	 */
	public static function singleton() {
		if(!isset(self::$instance)) {
			$c = __CLASS__;
			self::$instance = new $c;
		}

		return self::$instance;
	}

	/**
	 * @ignore
	 */
	public function __clone() {
		trigger_error('Enrober can not be cloned', E_USER_ERROR);
	}

	/**
	 * Returns the complete file system path to the cache file for the current version of the calling script.
	 * {@link set_theme()} must be called before get_cache_file_path() because the 
	 * theme name is used to generate the cache file's filename.
	 * <code>
	 * $enrober->set_theme();
	 * $cache_path = $enrober->get_cache_file_path();
	 * </code>
	 *
	 * @returns string The path to the cache file
	 */
	public function get_cache_file_path() {
		static $path = null;
		if ($path === null) {
			if ($this->theme_name === null) {
				trigger_error('Can not call $enrober->get_cache_file_path() without calling $enrober->set_theme() first', E_USER_ERROR);
			}
			$path = __DIR__ . '/cache/' . $this->get_cache_filename();
		}

		return $path;
	}

	/**
	 * Override the default theme.
	 *
	 * Can only be called once, subsequent calls will be ignored.  If no $theme is provided, the default
	 * theme is explicitly set as the theme.
	 *
	 * Certain methods (for example {@link is_cached} and {@link get_cache_file_path}) require that the theme be 
	 * set before they are called or they may raise warnings or errors, in these 
	 * cases call {@link set_theme}() (with no parameters if you want to use the default 
	 * theme) to avoid errors.
	 *
	 * If you are using set_theme(), it must be called before go(). go() uses the default theme if one has not been
	 * set before go() is called.
	 *
	 * <code>
	 * $enrober->set_theme('pressy'); // Change the theme from the default to the 'pressy' theme
	 * $enrober->go();
	 *
	 * // or
	 * $enrober->set_theme(); // Explicitly use the default theme.
	 * $enrober->go();
	 * </code>
	 *
	 * @param string $theme A string containing the name of the theme to wrap the page in.  If not provided the value of the ENROBER_DEFAULT_THEME constant from the config file is used.
	 * @since 1.1.0 Previously the theme could be set as a second parameter to {@link go()}, but it is now required that you use set_theme().
	 */
	public function set_theme($theme = null) {
		if ($this->theme_name === null) {
			if ($theme === null) {
				$theme = ENROBER_DEFAULT_THEME;
			}
			// Theme names can only contain a-z, A-Z, 0-9, underscore and dash to protect
			// against dir traversal in $theme in case users set the theme with
			// a var that can be manipulated by an attacker
			$theme = preg_replace('/[^a-z-_A-Z0-9]/', '', $theme);
		 	if (in_array($theme, $this->get_themes())) {
				$this->theme_name = $theme;
			} else {
				trigger_error('Enrober was unable to find the theme "' . $theme . '"', E_USER_ERROR);
			}
		}
	}

	/**
	 * Returns a list of themes currently available.
	 * <code>
	 * <?php
	 * echo '<ul>';
	 * // list currently installed themes
	 * foreach($enrober->get_themes() as $t) {
	 * 	echo "<li>$t</li>";
	 * }
	 * echo '</ul>';
	 * ?>
	 * </code>
	 * @returns array An array containing the names of all of the themes installed in this copy of Enrober
	 * @since 1.1.0
	 */
	public function get_themes() {
		static $themes = null;
		$out = array();

		if ($themes === null) {
			$d = dir(__DIR__ . '/themes/');
			while (($item = $d->read()) !== false) {
				if ($item !== '.' && $item !== '..' && is_dir($d->path . $item)) {
					$out[] = $item;
				}
			}
			$d->close();
		}
		return $out;
	}

	/**
	 * Returns the title of the page as it was set with go()
	 * Useful inside of XHR filters if transforming data into another format (JSON, XML, etc).
	 * @returns string
	 * @since 1.2.0
	 */
	public function get_title() {
		return $this->page_title;
	}

	/**
	 * Adds a cache modifier.
	 *
	 * Cache modifiers are used to differentiate between different versions of a page.
	 * This can be useful if you want to cache pages show different information when they
	 * have certain query string parameters, if you want to create different versions of the
	 * page based on the user agent, etc.
	 *
	 * @param string $modifier A string that identifies this version of the page. Can contain letters, numbers and underscores.
	 * <code>
	 * <?php
	 *	// In this example we'll build a store finder that takes a zipcode and 
	 *	// returns information for fictional stores in our database
	 *
	 *	// Get the zipcode the user entered
	 *	$zip = $_REQUEST['zip'];
	 *
	 *	// Strip out any non-numeric characters
	 *	$zip = preg_replace('/[^0-9]/', '', $zip);
	 *
	 *	// Call a function that does some extra validiation
	 *	if (is_valid_zip($zip)) {
	 *		// According to our rules it's a valid zipcode,
	 *		// instruct Enrober to cache a different version of
	 *		// the page for this zipcode
	 *		$enrober->add_cache_modifier('zip' . $zip);
	 *	}
	 *
	 *	// Start buffering the page with Enrober
	 *	$enrober->go("Stores in the $zip zipcode");
	 *
	 *
	 *
	 *	// Everything below here is for the sake of making the example complete 
	 *	// and has nothing to do with the use of add_cache_modifier()
	 *
	 *
	 *
	 *	if (is_valid_zip($zip)) {
	 *
	 *		echo "<h1>Stores in the $zip zipcode</h1>";
	 *
	 *		// Flush the h1 before the database query
	 *		ob_flush();
	 *		flush();
	 *
	 *		// Connect to our database
	 *		$mysqli = new mysqli('localhost', 'db_user', 'db_password', 'db_name');
	 *
	 *		// create a prepared statement to get the information from the database
	 *		if ($stmt = $mysqli->prepare('SELECT name, address, phone,' .
	 *				'hours FROM stores WHERE zip=?')) {
	 *
	 *				$stmt->bind_param('s', $zip);
	 *				$stmt->execute();
	 *
	 *				$stmt->bind_result($store_name, $store_address, $store_phone);
	 *
	 *				// Loop through the results and display them
	 *				while ($stmt->fetch()) {
	 *					echo '<div class="store">';
	 *					echo '<h2>' . htmlentities($store_name) . '</h2>';
	 *					echo '<ul>';
	 *					echo '<li>Address: ' . htmlentities($store_address) . '</li>';
	 *					echo '<li>Phone: ' . htmlentities($store_phone) . '</li>';
	 *					echo '</ul>';
	 *					echo '</div>';
	 *				}
	 *
	 *				$stmt->close();
	 *		}
	 *
	 *		$mysqli->close();
	 *	} else {
	 *		echo '<p>Sorry, no stores found in your area</p>';
	 *	}
	 * ?>
	 * </code>
	 */
	public function add_cache_modifier($modifier) {
		if (!in_array($modifier, $this->modifiers)) {
			// Strip any characters that are not alpha-numeric, or dash/underscore
			// since we will be using this as part of a filename, head off dir traversal dangers, etc.
			$modifier = preg_replace('/[^a-zA-Z0-9_]/', '', $modifier);
			$this->modifiers[] = $modifier;
		} else {
			trigger_error($modifier . ' not added to modifier list, already exists in list.', E_USER_WARNING);
		}
	}

	/**
	 * Returns the filename for caching this page
	 * @access private
	 * @returns string
	 */
	private function get_cache_filename() {
		static $name = null;
		if($name === null) {

			$mods = $this->modifiers;
			$enrober_mods = array();

			sort($mods);

			if ($this->theme_name === null) {
				trigger_error('$enrober->get_cache_filename() called before theme_name() set', E_USER_ERROR);
			} else {
				// If it's an XHR request, use a XHR token instead of a theme_name
				if (isset($_SERVER['HTTP_X_REQUESTED_WITH']) && strtolower($_SERVER['HTTP_X_REQUESTED_WITH']) === 'xmlhttprequest') {
					$name =  'XHR,';
				} elseif ($this->is_maintenance_mode && !$this->page_allowed_during_maintenance && !$this->is_maintenance_user) {
					$name = 'maintenance,' . $this->theme_name;
				} else {
					$name = $this->theme_name;
				}
			}

			if ($this->is_mobile() === true) {
				$enrober_mods[] =	'M';
			}

			$name .= ','; // , between theme and mods
			$name .= implode('-', $enrober_mods);
			$name .= ','; // , between Enrober mods and user mods
		 	$name .= implode('-', $mods) . ',' . base64_encode($_SERVER['SCRIPT_NAME']);
		}
		return $name;
	}

	/**
	 * Figure out if there is a cached version of this file/theme/cache-modifier combo.
	 *
	 * {@link set_theme}() must be called before is_cached() because it uses the theme name to figure out the name of the cache file
	 * @returns bool
	 */
	public function is_cached() {
			if ($this->theme_name === null) {
				trigger_error('$enrober->is_cached() called before theme_name() set', E_USER_ERROR);
			}
		return file_exists($this->get_cache_file_path());
	}

	/**
	 * Returns information regaurding whether we are currently in maintenance mode,
	 * whether the current user is the maintenance user and whether the current page
	 * is accessable during maintenance mode.
	 *
	 * If it is not maintenance mode, returns false.  If maintenance mode is currently
	 * active, an array containing the following is returned:
	 * <ul>
	 * <li>user - True if the current user has the ENROBER_MAINTENANCE_IP ip address, 
	 * false otherwise</li>
	 * <li>page - True if this page is allowed during maintenance, false if it is not.
	 * </ul>
	 *
	 * @returns false|array
	 * @since 1.2.0
	 */
	public function is_maintenance () {
		if ($this->is_maintenance_mode) {
			return array('user' => $this->is_maintenance_user,
				'page' => $this->page_allowed_during_maintenance);
		} else {
			return false;
		}
	}

	/**
	 * Determines if the current request is being served as a XHR request.
	 *
	 * @since 1.2.0
	 * @returns boolean
	 */
	function is_xhr () {
			if (ENROBER_X_REQUESTED_WITH === true && isset($_SERVER['HTTP_X_REQUESTED_WITH']) && strtolower($_SERVER['HTTP_X_REQUESTED_WITH']) === 'xmlhttprequest') {
				return true;
			} else {
				return false;
			}
	}

	/**
	 * Retrieves a cached copy of this file/theme combo
	 * @returns string The content of the cache file
	 */
	private function get_cache() {
		$file = file_get_contents($this->get_cache_file_path());
		$chunks = explode('<!--enrober>chunk-->', $file);

		return $chunks;
	}

	/**
	 * Manually specify whether or not the file's cache should be invalidated
	 *
	 * @param bool|int $valid - If True assume the cached is still good, if False the file is old and should be regenerated,
	 * if an integer is supplied it is assumed to be a unix timestamp and compared to the cache file's timestamp to determine
	 * the cache file's validity.
	 * <code>
	 * $enrober->manual_cache(true); // cache file is assumed to still be good no 
	 *                               // matter how old it is
	 * $enrober->manual_cache(false); // cache file is assumed to be expired and
	 *                                // must be re-processed
	 * $enrober->manual_cache(mktime(0,0,0,4,1)); // cache file is assumed to be
	 *                                            // expired if the current date is
	 *                                            // after midnight April 1st of the current year
	 * </code>
	 */
	public function manual_cache($valid) {
		// Just set to whatever value we get,
		// true/false/null/integers are handled in the if($this->is_cached()) section of go()
		if (is_int($valid)) {
			$this->cache_valid = intval($valid);
		} else if (is_bool($valid) || $valid === null) {
			$this->cache_valid = $valid;
		}
	}

 	/**
	 * Figures out the file system path to the current script
	 * @private
	 * @returns string
	 */
	private function get_script_path() {
		$path = realpath($_SERVER['SCRIPT_FILENAME']);
		// realpath can return a funky mixture of / and \ paths
		// make sure the entire path usues the correct DIRECTORY_SEPARATOR
		return preg_replace('/\/|\\\/', DIRECTORY_SEPARATOR, $path); // replaces / or \ with DIRECTORY_SEPARATOR
	}

	/**
	 * Adds <meta> keywords to the page.
	 * Multiple calls append to the list.
	 * @param mixed $keywords What to put in the <meta> keywords tag.  Either a string of comma delimited keywords or an array of keywords
	 *
	 * <code>
	 * $enrober->add_keywords('keyword1,keyword2');
	 * $enrober->add_keywords(array('keyword3', 'keyword4'));
	 * // the %keywords% template tag will now output
	 * // <meta name="keywords" content="keyword1,keyword2,keyword3,keyword4">
	 * </code>
	 */
	public function add_keywords($keywords) {
		if ($this->keywords !== '' && $keywords !== '') {
			$this->keywords .= ',';
		}

		if (is_array($keywords)) {
			$this->keywords .= implode(',', $keywords);
		} else {
			// assume it's a string
			$this->keywords .= $keywords;
		}
	}

	/**
	 * Adds a <meta> description to the page.
	 * Calling add_desc() a second time will replace any previous values.
	 *
	 * @param string $desc What to put in the <meta> description tag
	 * <code>
	 * $enrober->add_desc('This is a page description.');
	 * $enrober->add_desc('This description just replaced the old one.');
	 * </code>
	 */
	public function add_desc($desc) {
		$this->desc = $desc;
	}

	/**
	 * Examines the user agent string to try to determine whether the client is a mobile device.
	 *
	 * <code>
	 * $mobile = $enrober->m();
	 * var_dump($mobile);
	 *
	 * // When viewed on through a non-mobile browser, returns:
	 * // bool(false)
	 *
	 * // When viewed via Safari on an iPhone, returns:
	 * // array(8) {
	 * // 	["iphone"]=>
	 * // 	bool(true)
	 * // 	["android"]=>
	 * // 	bool(false)
	 * // 	["blackberry"]=>
	 * // 	bool(false)
	 * // 	["palm"]=>
	 * // 	bool(false)
	 * // 	["netfront"]=>
	 * // 	bool(false)
	 * // 	["safari"]=>
	 * // 	bool(true)
	 * // 	["opera"]=>
	 * // 	bool(false)
	 * // 	["windows"]=>
	 * // 	bool(false)
	 * // }
	 * </code>
	 *
	 * @since 1.1.0
	 * @returns mixed Contains information about mobile user agents.  If the user agent was not detected as mobile,
	 * it is set to false, otherwise it will be an array with the following keys set to true or false depending
	 * on whether or not the device appears to match the device type.
	 * <ul>
	 * <li>iphone: Device appears to be an iPhone</li>
	 * <li>android: Device appears to be an Android</li>
	 * <li>blackberry: Device appears to be an Black Berry</li>
	 * <li>palm: Device appears to be a Palm device</li>
	 * <li>windows: Device appears to be a Windows Phone or Windows CE device</li>
	 * <li>netfront: Device appears to be running the NetFront browser</li>
	 * <li>safari: Device appears to be running the Safari browser</li>
	 * <li>opera: Device appears to be running the Opera browser</li>
	 * </ul>
	 */
	public function m() {
		static $m = null;

		if ($m === null) {
			$ua = $_SERVER['HTTP_USER_AGENT'];
			if(!preg_match('/MIDP|Mobile|BlackBerry|webOS|^LG[-\/]|NetFront|UP.Browser|Windows (?:CE|Phone)|Palm/', $ua)) {
				$m = false;
			} else {
				$m = array('iphone' => false, 'android' => false, 'blackberry' => false, 'palm' => false, 'netfront' => false, 'safari' => false, 'opera' => false, 'windows' => false);
				// Device
				if(preg_match('/iPhone/', $ua)) {
					$m['iphone'] = true;
				} else if(preg_match('/android/i', $ua)) {
					$m['android'] = true;
				} else if(preg_match('/blackberry/i', $ua)) {
					$m['blackberry'] = true;
				} else if(preg_match('/Palm|webOS/', $ua)) {
					$m['palm'] = true;
				} else if(preg_match('/Windows/', $ua)) {
					$m['windows'] = true;
				}

				// Browsers
				if(preg_match('/netfront/i', $ua)) {
					$m['netfront'] = true;
				} else if(preg_match('/safari/i', $ua)) {
					$m['safari'] = true;
				} else if(preg_match('/opera/i', $ua)) {
					$m['opera'] = true;
				}
			}
		}

		return $m;
	}

	/**
	 * Determines if the page should be served as a mobile page.
	 *
	 * A return value of true does not necessarily mean that the user agent is a mobile device, 
	 * it is possible that {@link force_mobile()} has forced the mobile version.  For a more
	 * definitive answer use {@link m()}.
	 * 
	 * @returns bool
	 * @see m()
	 * @since 1.1.0
	 *
	 */
	public function is_mobile() {
		if (ENROBER_MOBILE_ENABLE) {
			// If we are not forcing mobile/non-mobile...
			if ($this->mobile === null) {
				// Detect mobile browsers
				if ($this->m() === false) {
					return false;
				} else {
					return true;
				}
			}
			// Forced, get value
			return $this->mobile;
		} else {
			// Automatic mobile detection is disabled...
			return false;
		}
	}

	/**
	 * Forces the mobile or non-mobile version of a page to be displayed.
	 *
	 * @param bool|null $force True === force mobile, False === force non-mobile, null === revert to auto-detect
	 * @since 1.1.0
	 */
	public function force_mobile($force) {
		switch ($force) {
		case true:
		case false:
		case null:
			$this->mobile = $force;
			break;
		}
	}

	/**
	 * Determines if a url should be accessable by a non-maintenance user during maintenance mode
	 */
	private function allowed_during_maintenance($url) {


		// Get list of patterns to allow
		if (file_exists(__DIR__ . '/cfg/maintenance-allow.txt')) {
			$allow_file = __DIR__ . '/cfg/maintenance-allow.txt';
		} else if (file_exists(__DIR__ . '/cfg/maintenance-allow.default.txt')) {
			$allow_file = __DIR__ . '/cfg/maintenance-allow.default.txt';
		} else {
			// No allow files at all? Everything is banned
			return false;
		}

		$allow = file($allow_file);

		// If there was an error reading the allow file, fallback to excluding everything
		if ($allow === false) {
			return false;
		}

		$allow_passed = false;
		foreach ($allow as $line) {
			$line = trim($line);
			// If this line is not blank or a comment
			if ($line !== '' && substr($line, 0, 1) !== '#') {
				$line = '/' . $line . '/';
				// if we matched an allow pattern
				if (preg_match($line, $url)) {
					// set flag that it matched and stop searching
					$allow_passed = true;
					break;
				}
			}
		}

		if ($allow_passed === false) {
			return false;
		}

		// Get list of patterns to disallow
		if (file_exists(__DIR__ . '/cfg/maintenance-disallow.txt')) {
			$disallow_file = __DIR__ . '/cfg/maintenance-disallow.txt';
		} else if (file_exists(__DIR__ . '/cfg/maintenance-disallow.default.txt')) {
			$disallow_file = __DIR__ . '/cfg/maintenance-disallow.default.txt';
		} else {
			// No allow files at all? Everything is banned
			return false;
		}

		$disallow = file($disallow_file);

		// If there was an error reading the disallow file, fallback to excluding everything
		if ($disallow === false) {
			return false;
		}

		foreach ($disallow as $line) {
			$line = trim($line);
			// If this line is not blank or a comment
			if ($line !== '' && substr($line, 0, 1) !== '#') {
				$line = '/' . $line . '/';
				// if we matched a disallow pattern
				if (preg_match($line, $url)) {
					return false;
				}
			}
		}

		// if it gets here without returning, it passed
		return true;
	}

	/**
	 * Invokes Enrober's page wrapping; invoke after calling any of the other Enrober methods you want to use and before page content starts.
	 *
	 * If you have not already added a {@link http://us.php.net/manual/en/function.ob-gzhandler.php ob_gzhandler} output buffer when go() is invoked, Enrober will automatically call ob_start('ob_gzhandler').
	 *
	 * <code>
	 * <?php
	 * $enrober->go(); // Invoke Enrober without a page title
	 * $enrober->go('Example'); // Invoke Enrober with the page title 'Example'
	 * ?>
	 * </code>
	 *
	 * @param string $title If specified this string will be inserted into a <title> tag
	 */
	public function go($title = '') {

		// Only start gzipping if it has not already begun there are cases (for 
		// example when using sessions with url rewriting enabled) when 
		// ob_gzhandler needs to be used before go() can be called.
		if (!in_array('ob_gzhandler', ob_list_handlers())) {
			ob_start("ob_gzhandler");
		}

		$this->page_title = $title;

		if ($this->theme_name === null) {
			$this->set_theme(ENROBER_DEFAULT_THEME);
		}

		$theme = $this->theme_name;

		if (defined('ENROBER_SITE_NAME') && ENROBER_SITE_NAME !== null) {
			$this->add_tag('site_name', ENROBER_SITE_NAME, 1);
		}

		if (defined('ENROBER_TAGLINE') && ENROBER_TAGLINE !== null) {
			$this->add_tag('tagline', ENROBER_TAGLINE, 1);
		}

		$is_maintenance_mode = $this->is_maintenance_mode;
		$is_maintenance_user = $this->is_maintenance_user;
		$page_allowed_during_maintenance = $this->page_allowed_during_maintenance;


		if ($is_maintenance_mode) {
			if ($is_maintenance_user) {
				// If we are the maintenance user, don't cache!
				$this->force_cache(false);
				$this->add_css(ENROBER_URL . '/css/maintenance.css'); 
			}
		}

		$desc = $this->desc;
		$keywords = $this->keywords;

		if ($is_maintenance_mode && !$page_allowed_during_maintenance && !$is_maintenance_user) {
			// no custom js/css loaded during maintenance since the normal page 
			// content is not loaded
			$js = '';
			$css = '';
		} else {
			// normal
			$js = $this->get_js();
			$css = $this->get_css();
		}

		$url = $this->url;

		// if force_cache() hasn't already overridden the config
		if($this->cache_pages === null) {
			if(!defined('ENROBER_CACHE')) {
				$this->cache_pages = false;
			} else {
				$this->cache_pages = ENROBER_CACHE;
			}
		}

		$theme_fs_path = __DIR__ . "/themes/$theme/";
		$theme_uri_path = ENROBER_URL . "/themes/$theme";

		// If we the client is a mobile UA and the current theme has a mobile version
		if ($this->is_mobile() && file_exists($theme_fs_path . 'm/') ) {
			// change theme paths to the mobile version
			$theme_fs_path .= 'm/';
			$theme_uri_path .= '/m';
		}

		if ($title !== '') {
			$title = '<title>' . (htmlentities($title)) . '</title>';
			$this->add_tag('title', $title, 1, false);
		}

		if ($desc !== '') {
			$this->add_tag('desc', '<meta name="description" content="' . htmlentities($desc) . '">', 1, false);
		}

		if ($keywords !== '') {
			$this->add_tag('keywords', '<meta name="keywords" content="' . htmlentities($keywords) . '">', 1, false);
		}

		if ($this->cache_pages === true) {
			$generated = 'Page cached ' . date(DATE_RFC822);
		} else {
			$generated = 'Page generated ' . date(DATE_RFC822);
		}

		if ($js !== '') {
			$this->add_tag('js', $js, 0, false);
		}

		if ($css !== '') {
			$this->add_tag('css', $css, 1, false);
		}

		if ($url !== null) {
			$this->add_tag('canonical', '<link rel="canonical" href="' . $url . '">', 1, false);
		}

		$this->add_tag('themepath', $theme_uri_path, 0, false);
		$this->add_tag('generated', $generated , 0);

		$tag_head_stack = $this->tag_stack_head;
		$tag_foot_stack = $this->tag_stack_foot;

		$filter_head_stack = &$this->filter_head_stack;
		$filter_foot_stack = &$this->filter_foot_stack;
		$filter_body_stack = &$this->filter_body_stack;

		$filter_everything_stack = &$this->filter_everything_stack;

		$filter_xhr_stack = &$this->filter_xhr_stack;

		$filter_always_head_stack = &$this->filter_always_head_stack;
		$filter_always_foot_stack = &$this->filter_always_foot_stack;
		$filter_always_body_stack = &$this->filter_always_body_stack;

		$theme_config = file_exists($theme_fs_path . 'config.json') ? json_decode(file_get_contents($theme_fs_path . 'config.json')) : array();

		$cache_pages = $this->cache_pages;
		$cache_file_path = $this->get_cache_file_path();

		// function that takes some text $data and applies each of the $filters
		$apply_filters = function(&$data, $filters) {
			// Apply each filter in the filters stack
			for($i = 0, $len = count($filters); $i < $len; $i++) {
				$new_data = call_user_func($filters[$i], $data);
				if ($new_data !== false) {
					$data = $new_data;
				}
			}
			return $data;
		};

		$get_header = function() use ($theme_config, $theme_fs_path, $tag_head_stack, &$filter_head_stack, $apply_filters) {
			static $header = null;

			if ($header === null) {
				// If head is defined in the config file use that or default
				$head_file = isset($theme_config->head) ? $theme_config->head : 'head.html';
				// Load header
				$header = file_get_contents($theme_fs_path . $head_file);

				// Process all of the %tags% for the head
				if(count($tag_head_stack) > 0) {
					foreach($tag_head_stack as $tag) {
						$header = str_replace('%' . $tag['name'] . '%', $tag['value'], $header);
					}
				}
				// Remove misc add_tag tags that were not processed
				$header = preg_replace('/%[^%\s]*%/', '', $header);

				$header = $apply_filters($header, $filter_head_stack);
			}

			return $header;
		};

		$cache_it = function($h, $b, $f) use ($cache_file_path, $cache_pages) {
			if($cache_pages === true) {
				file_put_contents($cache_file_path, $h . '<!--enrober>chunk-->' . $b . '<!--enrober>chunk-->' .  $f);
			}
		};

		// The callback must be an anonymous function so we can use closure to pass 
		// variables into the callback.
		//
		// Reminder: In PHP anonymous functions are early binding, pass variables
		// by reference if they are prone to change after the function's definition
		$enrobe_callback = function ($body) use ($get_header, $theme_fs_path,
			$tag_head_stack, $tag_foot_stack,
			&$filter_head_stack, &$filter_body_stack, &$filter_foot_stack, &$filter_xhr_stack,
			&$filter_everything_stack,
			&$filter_always_head_stack, &$filter_always_body_stack, &$filter_always_foot_stack,
			$theme_config,
			$page_allowed_during_maintenance, $is_maintenance_mode, $is_maintenance_user,
			$cache_pages, $cache_file_path,
			$apply_filters, $cache_it
		) {

			$body = $apply_filters($body, $filter_body_stack);

			// If X-requested-with support is turned on and this is an XHR request
			if (ENROBER_X_REQUESTED_WITH === true && isset($_SERVER['HTTP_X_REQUESTED_WITH']) && strtolower($_SERVER['HTTP_X_REQUESTED_WITH']) === 'xmlhttprequest') {
				// Set empty footer
				$header = ''; // Needs to be here since outside the callback $header is not set if it is XHR.
				$footer = '';
				$filter_always_foot_stack = array(); // ignore any foot filters since we don't have a foot
				$body = $apply_filters($body, $filter_xhr_stack);
			} else {
				$header = $get_header();
				// If is defined in the config file use that or default
				$foot_file = isset($theme_config->foot) ? $theme_config->foot : 'foot.html';
				// Non-XHR, load footer a usual
				$footer = file_get_contents($theme_fs_path . $foot_file);

				// Process all of the %tags% for the foot
				if(count($tag_foot_stack) > 0) {
					foreach($tag_foot_stack as $tag) {
						$footer = str_replace('%' . $tag['name'] . '%', $tag['value'], $footer);
					}
				}

				// Remove misc add_tag tags that were not processed
				$footer = preg_replace('/%[^%\s]*%/', '', $footer);

				$footer = $apply_filters($footer, $filter_foot_stack);
			}

			// Cache page
			if (count($filter_everything_stack) <= 0) {
				$cache_it($header, $body, $footer);
				$body = $apply_filters($body, $filter_always_body_stack);
				$footer = $apply_filters($footer, $filter_always_foot_stack);

				// Head has already been printed, output the body/foot
				return $body . $footer;
			} else {
				// Return it with chunkers in it so that always filters can be run after the everything filters are applied
				return $header . '<!--enrober>chunk-->' . $body . '<!--enrober>chunk-->' . $footer;
			}
		}; // Anon function, the ; is needed  here!

		$valid_cache = false;
		if($this->is_cached()) {
			if($this->cache_valid === null) {
				// automatic cache expire based on source file's modified date
				if (filemtime($this->get_script_path()) < filemtime($cache_file_path)) {
					$valid_cache = true;
				}
			} else if(is_int($this->cache_valid)) {
				// expire based on a specified time
				// if the current time is smaller than the expire time, cache is valid
				if ($this->cache_valid > time()) {
					$valid_cache = true;
				}
			} else {
				// cache validity was set manually to true or false
				$valid_cache = $this->cache_valid;
			}
		} else {
			// no cached file, can't be valid!
			$valid_cache = false;
		}

		if ($is_maintenance_mode) {
			// we are not the maintenance user
			if (!$is_maintenance_user) {
				// this is page is not allowed in maintenance mode
				if (!$page_allowed_during_maintenance) {
					/* Send a 503 so Google and other search engines don't mis-index this as the page
					 * and know to check back later for the real content */
					header('HTTP/1.1 503 Service Temporarily Unavailable');
					header('Retry-After: ' . (is_numeric(ENROBER_MAINTENANCE_DONE_AT) ? ENROBER_MAINTENANCE_DONE_AT : date(DATE_RFC1123, strtotime(ENROBER_MAINTENANCE_DONE_AT))));
				} // else {} do nothing, it's allowed in maintenance mode
			} else {
				// We are the maintenance user
				// Insert a big annoying MAINTENANCE banner on all pages so you realize you are still in maintenance mode
				// and don't leave it when you are done
				$this->add_filter(function ($body) use ($page_allowed_during_maintenance) {
					return '<div class="maintenance"><p>MAINTENANCE MODE</p><p>This page is currently' .
						($page_allowed_during_maintenance === true ? ' ' : ' NOT ') . 'accessible to the public.</p></div>' . $body;
				}, ENROBER_FILTER_BODY);
			}
		}

		if ($this->cache_pages && $valid_cache) {
			// Display the cached file

			$cache = $this->get_cache();

			echo $apply_filters($cache[0], $this->filter_always_head_stack);
			ob_flush();
			flush();
			echo $apply_filters($cache[1], $this->filter_always_body_stack);
			ob_flush();
			flush();
			echo $apply_filters($cache[2], $this->filter_always_foot_stack);
			ob_flush();
			flush();

			while (@ob_end_flush()); // flush any buffers that are still around
			exit;
		} else {

			// Maintenance mode is off or this page is allowed during maintenance, 
			// proceed as usual.

			// If it's not an XHR request
			if (!(ENROBER_X_REQUESTED_WITH === true && isset($_SERVER['HTTP_X_REQUESTED_WITH']) && strtolower($_SERVER['HTTP_X_REQUESTED_WITH']) === 'xmlhttprequest')) {

				// If there are any filter everything filters defined start an output buffer to process them.
				// This affects the page content, the header and footer all at once.  Since it has to
				// buffer the whole page it prevents early head flushing!!  Use with caution.
				if (count($this->filter_everything_stack) > 0 && !($is_maintenance_mode && !$page_allowed_during_maintenance && !$is_maintenance_user)) {
					$filter_everything_stack = &$this->filter_everything_stack;
					$filter_evertying = function($data) use ($apply_filters, $cache_it, &$filter_everything_stack, &$filter_always_head_stack, &$filter_always_body_stack, &$filter_always_foot_stack)
					{
						$data = $apply_filters($data, $filter_everything_stack);
						$chunks = explode('<!--enrober>chunk-->', $data);

						$cache_it($chunks[0], $chunks[1], $chunks[2]);

						$header = $apply_filters($chunks[0], $filter_always_head_stack);
						$body = $apply_filters($chunks[1], $filter_always_body_stack);
						$footer = $apply_filters($chunks[2], $filter_always_foot_stack);

						return $header . $body . $footer;
					}; // Anon function, the ; is needed
					ob_start($filter_evertying);
				} else {
					$header = $get_header();

					$header = $apply_filters($header, $filter_always_head_stack);

					// Flush the head early
					echo $header;
					ob_flush();
					flush();
				}

				// If we are in maintenance mode and this page is not allowed durning maintenance mode
				// and we are not the maintenance user
				if ($is_maintenance_mode && !$page_allowed_during_maintenance && !$is_maintenance_user) {
					/* Get our maintenance page */
					if (file_exists(__DIR__ . '/cfg/maintenance.html')) {
						$body = file_get_contents(__DIR__ . '/cfg/maintenance.html');
					} else if (file_exists(__dir__ . '/cfg/maintenance.default.html')) {
						$body = file_get_contents(__dir__ . '/cfg/maintenance.default.html');
					} else {
						$body = '<h1>Maintenance Mode</h1><p>Sorry, this page is currently unavailable.</p>';
					}

					if (is_numeric(ENROBER_MAINTENANCE_DONE_AT)) {
						$maint_done_at = strtotime('+' . ENROBER_MAINTENANCE_DONE_AT . ' seconds');
					} else {
						$maint_done_at = strtotime(ENROBER_MAINTENANCE_DONE_AT);
					}
					$maint_done_at = date('g:i:s a \o\n M j, Y', $maint_done_at);
					$body = str_replace('%maintenance_end%', $maint_done_at, $body);

					// wrap maintenance page in the theme
					echo $enrobe_callback($body);
					while (@ob_end_flush()); // flush any buffers
					exit; // no need to go on and proccess the calling script
				}
			}

			ob_start($enrobe_callback); 
		}
		// Function ends, normal page logic takes over; ob_ callbacks end when page ends
	}

	/**
	 * Adds a CSS file to be inserted at the %css% tag.
	 * If a stylesheet is added a second time it is ignored.
	 *
	 * <code>
	 * <?php
	 * // Linked as a regular stylesheet
	 * $enrober->add_css('extra.css');
	 *
	 * // Linked as a print-only stylesheet
	 * $enrober->add_css('print.css', array('media' => 'print');
	 *
	 * // Linked as an alternative stylesheet
	 * $enrober->add_css('alternate.css', array('rel' => 'alt', 'title' => 'Alternate layout');
	 *
	 * // Not added because it is a duplicate
	 * $enrober->add_css('extra.css');
	 * ?>
	 * </code>
	 *
	 * @param string $filepath The url to the css file to be included.  If the css file is in the same directory as the calling script only the filename needs to be specified.
	 * @param array $config An array that allows you to add additional optional attributes to the link tag.
	 * These attributes are:
	 * <ul>
	 * <li>'media': Which media types the stylesheet should be applied to, 'all' by default.</li>
	 * <li>'rel': The rel attribute of the <link> tag, 'stylesheet' by default can also be set to 'alternate stylesheet' to specify an alternate stylesheet.  If set to 'alt' it will automatically be expanded to 'alternate stylesheet'.</li>
	 * <li>'title': The title attribute for the link, most browsers that support alternative stylesheets will use this as a label in the stylesheet switching interface.</li>
	 * </ul>
	 *
	 */
	public function add_css($filepath, $config = null) {

		static $files = array();

		if (!is_array($config)) {
			$config = array();
		}

		if (array_key_exists('rel', $config)) {
			// Allow $rel to be set to 'alt' as a shortcut for 'alternate stylesheet'
			if ($config['rel'] === 'alt') {
				$config['rel'] = 'alternate stylesheet';
			}
		} else {
			$config['rel'] = 'stylesheet';
		}

		if (!array_key_exists('media', $config)) {
			$config['media'] = 'all';
		}

		if (!array_key_exists('title', $config)) {
			$config['title'] = 'Stylesheet #' . count($this->css_stack) + 1;
		}

		$new_item = array('file' => $filepath, 'media' => $config['media'], 'rel' => $config['rel'], 'title' => $config['title']);

		/* Don't add if it's already in the stack */
		if (!in_array($filepath, $files)) {
			$this->css_stack[] = $new_item;
			$files[] = $filepath;
		}
	}

	/**
	 * Adds a JavaScript file to be linked via the %js% tag.
	 * If a script is added a second time it is ignored.
	 * 
	 * @param string $filepath The url of the JavaScript file to add absolute or relative to the calling script.
	 *
	 * <code>
	 * <?php
	 * $enrober->add_js('http://cdn.example.com/1.5/jquery.min.js'); // Import jQuery from a CDN
	 * $enrober->add_js('blink.js'); // Add a JavaScript file that is in the same
	 *                               // directory as the page using it.
	 * ?>
	 * </code>
	 */
	public function add_js($filepath) {
		/* Don't add if it's already in the stack */
		if (!in_array($filepath, $this->js_stack)) {
			$this->js_stack[] = $filepath;
		}
	}

	/**
	 * Adds a content filter to be applied with ob_start()
	 *
	 * @param string $filter A string with the name of the filter function to use
	 * @param bool $filter_mode What part of the page should be filtered?
	 * <ul>
	 * <li>ENROBER_FILTER_BODY - Filter the body of the page.  If no filter mode is specified, this is used as the default.</li>
	 * <li>ENROBER_FILTER_HEAD - Filter the header of the template.</li>
	 * <li>ENROBER_FILTER_FOOT - Filter the footer of the template.</li>
	 * <li>ENROBER_FILTER_ALL - Adds a head, body and foot filter with one command.</li>
	 * <li>ENROBER_FILTER_EVERYTHING - Buffers the entire page and then applies your filter to the whole thing at once.</li>
	 * <li>ENROBER_FILTER_ALWAYS_HEAD - Filter the  head.  Happens at request time, even on cached pages.</li>
	 * <li>ENROBER_FILTER_ALWAYS_BODY - Filter the  body.  Happens at request time, even on cached pages.</li>
	 * <li>ENROBER_FILTER_ALWAYS_FOOT - Filter the  foot.  Happens at request time, even on cached pages.</li>
	 * <li>ENROBER_FILTER_ALWAYS_ALL - Adds a head, body and foot filter with one command.  Happens at request time, even on cached pages.</li>
	 * <li>ENROBER_FILTER_XHR - Before an XHR request is returned it is run through any XHR filters.  Ideal for returning XHR requests in different formats (JSON, XML, etc.)</li>
	 * <li>false - Synonym for ENROBER_FILTER_BODY (deprecated)</li>
	 * <li>true - Synonym for ENROBER_FILTER_ALL (deprecated)</li>
	 * </ul>
	 *
	 * Note: If a request is an XHR request, it is filtered with any body filters, all filters, XHR filters and always filters.  Evertything filters are not run on XHR requests.  Also "always" filters are always run last so care must be taken that you don't corrupt any data transformations done by a XHR filter, for instance insert non-escaped characters into JSON, or search and replace tag names or property names in XML, etc.
	 *
	 * @since 1.2.0 More granularity in where filters are applied. True/false have been deprecated.
	 *
	 * <code>
	 * <?php
	 * // Apply the HTML Tidy callback to the body of the
	 * // page to fix any broken markup (Requires the Tidy
	 * // extension for PHP).  The entire document is processed as one piece.
	 * $enrober->add_filter('ob_tidyhandler', ENROBER_FILTER_EVERYTHING);
	 *
	 * // This filter function collapses any whitespace
	 * // between HTML tags down to a single space
	 * function collapse_whitespace($in) {
	 * 	return preg_replace('/>\s+</','> <', $in); 
	 * }
	 *
	 * // add our custom filter, filtering everything
	 * // including the header and footer.  Each section is filtered by itself.
	 * $enrober->add_filter('collapse_whitespace', ENROBER_FILTER_ALL);
	 *
	 * // Censor the words foo, bar and baz.
	 * function censor($in) {
	 * 	$out = str_replace('foo', '***', $in);
	 * 	$out = str_replace('bar', '***', $out);
	 * 	$out = str_replace('baz', '***', $out);
	 * 	return $out;
	 * }
	 *
	 * $enrober->add_filter('censor'); // Implicitly only filtering the body of the page.
	 * ?>
	 * </code>
	 */
	public function add_filter($filter, $filter_mode = ENROBER_FILTER_BODY) {

		// Don't allow gzipping since we already do that automatically
		if ($filter === 'ob_gzhandler') {
			trigger_error("Can not add ob_gzhandler as a filter, Enrober automatically wraps your page with it already.", E_USER_WARNING);
			return;
		}	

		if ($filter_mode === ENROBER_FILTER_ALL) {
			$this->filter_head_stack[] = $filter;
			$this->filter_body_stack[] = $filter;
			$this->filter_foot_stack[] = $filter;
		} elseif ($filter_mode === ENROBER_FILTER_BODY) {
			$this->filter_body_stack[] = $filter;
		} elseif ($filter_mode === ENROBER_FILTER_HEAD) {
			$this->filter_head_stack[] = $filter;
		} elseif ($filter_mode === ENROBER_FILTER_FOOT) {
			$this->filter_foot_stack[] = $filter;
		} elseif ($filter_mode === ENROBER_FILTER_XHR) {
			$this->filter_xhr_stack[] = $filter;
		} elseif ($filter_mode === ENROBER_FILTER_ALWAYS_HEAD) {
			$this->filter_always_head_stack[] = $filter;
		} elseif ($filter_mode === ENROBER_FILTER_ALWAYS_BODY) {
			$this->filter_always_body_stack[] = $filter;
		} elseif ($filter_mode === ENROBER_FILTER_ALWAYS_FOOT) {
			$this->filter_always_foot_stack[] = $filter;
		} elseif ($filter_mode === ENROBER_FILTER_ALWAYS_ALL) {
			$this->filter_always_head_stack[] = $filter;
			$this->filter_always_body_stack[] = $filter;
			$this->filter_always_foot_stack[] = $filter;
		} elseif ($filter_mode === false) {
			$this->filter_body_stack[] = $filter;
		} else if ($filter_mode === ENROBER_FILTER_EVERYTHING || $filter_mode === true) {
			$this->filter_everything_stack[] = $filter;
		} else {
		}
	}

	/**
	 * Adds a new theme tag.
	 *
	 *
	 * Theme tags allow you to specify where things Enrober generates are injected into the theme. There are a number of built in tags:
	 * <ul>
	 * <li>%title% - The page's <title> tag</li>
	 * <li>%desc% - The page's description <meta> tag</li>
	 * <li>%keywords% - The page's keywords <meta> tag</li>
	 * <li>%themepath% - The URI for the directory that the current theme is stored in, useful for including the theme's CSS file.</li>
	 * <li>%css% - Any <link> tags for CSS files added via $enrober->add_css()</li>
	 * <li>%js% - Any <script> tags for JavaScript files added via $enrober->add_js()</li>
	 * <li>%generated% - Inserts a "Generated on" or "Cached on" message depending on whether or not caching is enabled on the current page.</li>
	 * <li>%canonical% - Where in the head to insert the Canonical URL if one was set with $enrober->url()</li>
	 * <li>%site_name% - The value for ENROBER_SITE_NAME specified in enrober-config.php</li>
	 * <li>%tagline% - The value for ENROBER_TAGLINE specified in enrober-config.php</li>
	 * </ul>
	 * 
	 * If you add your own tags to your templates you can then set their value for any particular page with add_tag().
	 * 
	 *
	 * <code>
	 * <?php
	 * $enrober->add_tag('foo', 'bar'); // Adds the text "bar" anywhere in the 
	 *                                  // header or footer that "%foo%" appears
	 * $enrober->add_tag('foo', 'bar', 0); // Same as previous
	 * $enrober->add_tag('foo', 'bar', 1); // Adds the text "bar" only in the header
	 * $enrober->add_tag('foo', 'bar', 2); // Adds the text "bar" only in the footer
	 *
	 * // By default anything added to a theme tag is run through htmlentities()
	 * // to prevent XSS attacks. If you need to use HTML you can set the $sanitize
	 * // parameter to false but if you are including any user generated content
	 * // in it you will need to sanitize it yourself!
	 * $enrober->add_tag('foo', '<em>bar</em>'); // Adds the text "bar", wrapped
	 *                                           // in a text representation of em tags:
	 *                                           // "&lt;em&gt;bar&lt;/em&gt;"
	 * $enrober->add_tag('foo', '<em>bar</em>', 0, true); // Same as previous
	 * $enrober->add_tag('foo', '<em>bar</em>', 0, false); // Adds the text "bar",
	 *                                                     // wrapped in em tags:
	 *                                                     // "<em>bar</em>"
	 * ?>
	 * </code>
	 *
	 * @param string $name The name of the theme tag to define (sans %%)
	 * @param string $value The value to insert at the new theme tag
	 * @param integer $location Where to process the tag.  0 = Header and Footer, 1 = Header only, 2 = Footer only
	 * @param bool $sanitize If true, the $value is run through htmlentities() before inclusion in the page. Set to false if you need to include HTML, but you will need to sanitize any user input you include yourself!
	 */
	public function add_tag($name, $value, $location = 0, $sanitize = true) {

		$name = preg_replace('/^%|%$/', '', $name);

		if ($sanitize !== false) {
			$value = htmlentities($value);
		}

		$new_item = array('name' => $name, 'value' => $value);

		if ($location === 0 || $location === 1) {
			$this->tag_stack_head[] = $new_item;
		}
		if ($location === 0 || $location === 2) {
			$this->tag_stack_foot[] = $new_item;
		}
	}

	/**
	 * Sets/gets the canonical URL for the page to be inserted at the %canonical% tag.
	 *
	 * If called with no arguments or with null, returns the current canonical url. 
	 * If an empty string, removes the canonical url.  If it is anything else it 
	 * is set as the canonical url.
	 *
	 * Google's webmaster central has more information about {@link http://www.google.com/support/webmasters/bin/answer.py?answer=139066 canonical URLs}.
	 *
	 * <code>
	 * $enrober->url('index.php');
	 * // or
	 * $enrober->url('/');
	 * // or
	 * $enrober->url('http://www.example.com/');
	 * </code>
	 *
	 * @param string $url The url to set as the canonical url.
	 */
	function url($url = null) {
		if ($url === null) {
			return $this->url;
		} else if ($url === '') {
			$this->url = null;
		} else {
			$this->url = $url;
		}
	}

	/**
	 * Manually turn on/off caching, overriding the ENROBER_CACHE constant from the config file.
	 *
	 * When caching is turned on with force_cache(), normal caching rules apply.  To guarantee that 
	 * the cached version of a page gets returned when one exists use {@link manual_cache()} set to true.
	 *
	 * @param bool $cache Should caching be enabled for this page?  True = Cache, False = Don't cache
	 */
	public function force_cache($cache) {
		if($cache === true) {
			$this->cache_pages = true;
		} else {
			$this->cache_pages = false;
		}
	}

	/**
	 * Returns a list of <link> tags for each CSS file in the css stack.
	 * @returns string
	 */
	private function get_css() {
		$out = '';
		foreach($this->css_stack as $stylesheet) {
			$out .= '<link href="' . $stylesheet['file'] . '" rel="' . $stylesheet['rel'] . '" type="text/css" media="' . $stylesheet['media'] . '" title="' . $stylesheet['title'] . '">';
		}
		return $out;
	}

	/**
	 * Returns a string containing <script> tags to add to the document via the %js% tag
	 * @returns string
	 */
	private function get_js() {
		$out = '';
		foreach($this->js_stack as $js_file) {
			$out .= '<script src="' . $js_file . '" type="text/javascript"></script>';
		}
		return $out;
	}
}
// Create our $enrober object for scripts to invoke
$enrober = Enrober::singleton();
?>
